/*****************************************************************************
 *                                                                           *
 * Copyright 2006 Timothy F Clarke (timothyclarke@gmail.com)                 *
 *                                                                           *
 * This work is licensed under the Creative Commons Attribution License.     *
 *                                                                           *
 * To view a copy of this license, visit                                     *
 * http://creativecommons.org/licenses/by/2.0 or send a letter to Creative   *
 * Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.          *
 *                                                                           *
 * You are free:                                                             *
 *     - to copy, distribute, display, and perform the work                  *
 *     - to make derivative works                                            *
 *     - to make commercial use of the work                                  *
 * Under the following conditions:                                           *
 *     - Attribution. You must give the original author credit.              *
 *     - For any reuse or distribution, you must make clear to others the    *
 *           license terms of this work.                                     *
 *     - Any of these conditions can be waived if you get permission from    *
 *           the copyright holder.                                           *
 * Your fair use and other rights are in no way affected by the above.       *
 *                                                                           *
 * Notification of use of this work would be appreciated but is in no way    *
 * mandatory.                                                                *
 *                                                                           *
 ****************************************************************************/



#include "mymaths.inc"



// Increasing this value will make the straight stem straighter
#declare STRAIGHT_STEM_FRACTION = 16;



/*
 * Construct the shape of stem which may be used for (as an example) a meadow
 * foxtail.  Note it isn't truly straight but has a slight bend in it as would
 * be found in nature.  The top may be found by calling
 * StraightStemTipLocation, or alternatively any location on the stem may be
 * found by calling StraightStemLocation.
 *
 * i_radius: the _approximate_ radius of the stem, which will vary to a slight
 *     degree as we travel along the curve of the stem
 * i_length: the length of the stem travelling from the base to the to
 *     _in_a_curved_line_ (i.e. including the curve)
 */
#macro MakeStraightStem( i_radius, i_length )
    #local circumference = ( i_length * STRAIGHT_STEM_FRACTION );
    #local torus_major_radius = (( circumference / pi ) / 2 );

    difference
    {
        torus
        {
            torus_major_radius, i_radius
            rotate <-90, 0, 0>
        }
        box
        {
            // remove the lower half of the torus
            <-( torus_major_radius + i_radius ),
                -( torus_major_radius + i_radius ), -i_radius>
            <( torus_major_radius + i_radius ), 0, i_radius>
        }
        box
        {
            // remove the upper right quarter of the torus
            <0, 0, -i_radius>
            <( torus_major_radius + i_radius ),
                ( torus_major_radius + i_radius ), i_radius>
        }
        box
        {
            // remove the remainder of the torus leaving us with just one
            // STRAIGHT_STEM_FRACTIONth of the original
            <0, 0, -i_radius>
            <( torus_major_radius + i_radius ),
                ( torus_major_radius + i_radius ), i_radius>
            rotate <0, 0, 90 - ( 360 / STRAIGHT_STEM_FRACTION )>
        }

        // move the stem to the 0 position
        translate <torus_major_radius, 0, 0>
    }
#end


/*
 * Find the rectangular coordinates of a straight stem using polar coordinates.
 *
 * i_length: the length of the stem
 * i_radians: an angle in radians from which to project the rectangular
 *     coordinates
 */
#macro StraightStemLocation( i_length, i_radians )
    #local circumference = ( i_length * STRAIGHT_STEM_FRACTION );
    #local torus_major_radius = (( circumference / pi ) / 2 );

    <torus_major_radius - ( torus_major_radius * cos( i_radians )),
        ( torus_major_radius * sin( i_radians )), 0>
#end


/*
 * Find the rectangular coordinates of a straight stem tip.
 *
 * i_length: the length of the stem
 */
#macro StraightStemTipLocation( i_length )
    StraightStemLocation( i_length, Radians( 360/STRAIGHT_STEM_FRACTION ))
#end


/*
 * Construct the shape of stem which may be used for (as an example) a poppy.
 * The stem is bent back upon itself so the top may be found at
 * <0, i_length, 0>.
 *
 * i_radius: the _approximate_ radius of the stem, which will vary to a slight
 *     degree as we travel along the curve of the stem
 * i_length: the length of the stem travelling from the base to the to
 *     _in_a_straight_line_ (i.e. not including the curve)
 */
#macro MakeCurvedStem( i_radius, i_length )
    difference
    {
        torus
        {
            i_length, i_radius
            rotate <-90, 0, 0>
        }
        box
        {
            // remove the lower half of the torus
            <-( i_length + i_radius ), -( i_length + i_radius ), -i_radius >
            <( i_length + i_radius ), 0, i_radius>
        }
        box
        {
            // remove the upper right quarter of the torus
            <0, 0, -i_radius>
            <( i_length + i_radius ), ( i_length + i_radius ), i_radius>
        }

        translate <i_length, 0, 0>
        scale <0.25, 1, 1>

        // skew the stem so that the top is aligned directly above the base
        matrix <+1.00, +0.00, +0.00,
                -0.25, +1.00, +0.00,
                +0.00, +0.00, +1.00,
                +0.00, +0.00, +0.00>
    }
#end


/*
 * Make a leaf blade.
 *
 * i_rnd_seed: an initialized random number seed
 * i_segment_height: the height of the each group of four triangles making up
 *     the blade.  The smaller the height the smoother the blade curves will
 *     appear, but the larger the mesh will be
 * i_length: the length of blade to create
 * i_width: the width of blade to create
 * i_master_curve_speed: controls the number of segments created before the
 *     segment direction is changed.  A lower speed will make the curve
 *     appear smoother, but the curve will also be shallower
 * i_length_curve_speed: the speed the leaf blade will curve away on the z
 *     axis
 * i_pre_curve_speed: the length of blade which will continue in an upward
 *     trend before dropping down again
 * i_vertical_curve_speed: the speed the curve will move down once the
 *     downward speed has begun
 */
#macro MakeLeafBlade( i_rnd_seed, i_segment_height, i_length, i_width,
    i_master_curve_speed, i_length_curve_speed, i_pre_curve_length,
    i_vertical_curve_speed )

    #local ULP_OFFSET = 0;
    #local UMP_OFFSET = 1;
    #local URP_OFFSET = 2;
    #local NUM_POINTS = 3;

    #local num_segments =
        floor( i_length / i_segment_height ) +
            ( mod( i_length, i_segment_height ) ? NUM_POINTS : 0 );

    #local points = array[ NUM_POINTS * num_segments ]

    /*----------------------------------------------------------------------*
     * Find the points for the lower triangles which start at y=0.  We can  *
     * then build on these points by progressivly building up point y.      *
     * Note that we don't store the lower points as these are the only      *
     * ones we need to calculate, the others can be copied from the upper   *
     * points.                                                              *
     *                                                                      *
     *                    upper middle point (ump)                          *
     *                             |                                        *
     * upper left point (ulp) o----o----o upper right point (urp)           *
     *                        |   /|\   |                                   *
     *                        |  / | \  |                                   *
     *                        | /  |  \ |                                   *
     *                        |/   |   \|                                   *
     * lower left point (llp) o----o----o lower right point (lrp)           *
     *                             |                                        *
     *                     lower middle point (lmp)                         *
     *----------------------------------------------------------------------*/

    // lower left point
    #local llp = <( -i_width * 0.5 ), 0, 0>;

    // lower middle point
    #local lmp = <0, 0, ( i_width * 0.15 )>;

    // lower right point
    #local lrp = <abs( llp.x ), llp.y, llp.z>;

    // upper left point
    #local ulp = <llp.x, ( llp.y + i_segment_height ), llp.z>;
    #local points[ ULP_OFFSET ] = ulp;

    // upper middle point
    #local ump = <lmp.x, ( lmp.y + i_segment_height ), lmp.z>;
    #local points[ UMP_OFFSET ] = ump;

    // upper right point
    #local urp = <lrp.x, ( lrp.y + i_segment_height ), lrp.z>;
    #local points[ URP_OFFSET ] = urp;

    /*----------------------------------------------------------------------*
     * Use the points defined above to find points for our remaining        *
     * triangles.  All that need be done is to progressivly build up point  *
     * y, and apply the curve as requested by the client.                   *
     *----------------------------------------------------------------------*/

    #local vertical_curve = 0;
    #local length_curve = i_length_curve_speed;
    #local bend_extent = floor( num_segments * i_pre_curve_length );
    #local segment_idx = 1;

    #while ( bend_extent < num_segments )
        #local bend_extent =
            min(( bend_extent + floor( num_segments * i_master_curve_speed )),
                num_segments );

        #while ( segment_idx < bend_extent )
            #local ulp =
                <ulp.x, ( ulp.y + i_segment_height - vertical_curve ),
                    ( ulp.z + length_curve )>;
            #local points[( segment_idx * NUM_POINTS ) + ULP_OFFSET ] = ulp;

            #local ump =
                <ump.x, ( ump.y + i_segment_height - vertical_curve ),
                    ( ump.z + length_curve )>;
            #local points[( segment_idx * NUM_POINTS ) + UMP_OFFSET ] = ump;

            #local urp =
                <urp.x, ( urp.y + i_segment_height - vertical_curve ),
                    ( urp.z + length_curve )>;
            #local points[( segment_idx * NUM_POINTS ) + URP_OFFSET ] = urp;

            #local segment_idx = ( segment_idx + 1 );
        #end

        #local vertical_curve = ( vertical_curve + i_vertical_curve_speed );
        #local length_curve = ( length_curve + i_length_curve_speed );
    #end

    /*----------------------------------------------------------------------*
     * All the points have been calculated, now do the actual triangle      *
     * construction.                                                        *
     *----------------------------------------------------------------------*/

    object
    {
        mesh
        {
            #local segment_idx = 0;

            #while ( segment_idx < ( num_segments - 1 ))
                #local ulp =
                    points[( segment_idx * NUM_POINTS ) + ULP_OFFSET ];
                #local ump =
                    points[( segment_idx * NUM_POINTS ) + UMP_OFFSET ];
                #local urp =
                    points[( segment_idx * NUM_POINTS ) + URP_OFFSET ];

                triangle { llp, lmp, ump }
                triangle { lmp, lrp, ump }
                triangle { llp, ulp, ump }
                triangle { lrp, urp, ump }

                #local llp = ulp;
                #local lmp = ump;
                #local lrp = lrp;

                #local segment_idx = ( segment_idx + 1 );
            #end

            #local llp = ulp;
            #local lmp = ump;
            #local lrp = lrp;

            #local ump = points[( segment_idx * NUM_POINTS ) + UMP_OFFSET ];

            triangle { llp, lmp, ump }
            triangle { lmp, lrp, ump }
        }
    }
#end
