/*****************************************************************************
 *                                                                           *
 * Copyright 2006 Timothy F Clarke (timothyclarke@gmail.com)                 *
 *                                                                           *
 * This work is licensed under the Creative Commons Attribution License.     *
 *                                                                           *
 * To view a copy of this license, visit                                     *
 * http://creativecommons.org/licenses/by/2.0 or send a letter to Creative   *
 * Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.          *
 *                                                                           *
 * You are free:                                                             *
 *     - to copy, distribute, display, and perform the work                  *
 *     - to make derivative works                                            *
 *     - to make commercial use of the work                                  *
 * Under the following conditions:                                           *
 *     - Attribution. You must give the original author credit.              *
 *     - For any reuse or distribution, you must make clear to others the    *
 *           license terms of this work.                                     *
 *     - Any of these conditions can be waived if you get permission from    *
 *           the copyright holder.                                           *
 * Your fair use and other rights are in no way affected by the above.       *
 *                                                                           *
 * Notification of use of this work would be appreciated but is in no way    *
 * mandatory.                                                                *
 *                                                                           *
 ****************************************************************************/



#include "mycsg.inc"
#include "mymaths.inc"

#include "generic_botanical.inc"



#ifndef ( RENDER_INFLORESCENCE )
    // Set this value to false to speed rendering by not creating the
    // inflorescent
    #declare RENDER_INFLORESCENCE = true;
#end

#ifndef ( RENDER_SPIKELETS )
    // Set this value to false to speed rendering by not creating the
    // inflorescent spikelets
    #declare RENDER_SPIKELETS = true;
#end

#ifndef ( RENDER_LEAVES )
    // Set this value to false to speed rendering by not creating the
    // leaves
    #declare RENDER_LEAVES = true;
#end



/******************************************************************************
 ******************************************************************************
 **                                                                          **
 ** The Meadow Foxtail (Alopecurus pratensis), also known as the Field       **
 ** Meadow Foxtail, is a perennial grass, belonging to the grass family      **
 ** (Poaceae).  It is native to Europe and Asia.                             **
 **                                                                          **
 ** This common plant is found on grasslands, especially on neutral soils.   **
 ** It can grow to a height of about 110 cm.  The stem is erect or           **
 ** geniculate at the base.  The leaves are about 5 mm wide.  Meadow Foxtail **
 ** has a cylindrical inflorescence, with glumes about 5-10 mm wide and      **
 ** spikelets about 4-6 mm long.  It flowers from late spring to early       **
 ** summer and often again from late summer till autumn.                     **
 **                                                                          **
 ** This species is widely cultivated for pasture and hay, and has become    **
 ** naturalised in many areas outside of its native range, including         **
 ** Australia and North America.                                             **
 **     - http://en.wikipedia.org/wiki/Alopecurus_pratensis                  **
 **                                                                          **
 ** Flower head forms a soft, compact cylindrical spike up to 9cm long       **
 **     - http://www.naturedetectives.org.uk/identify/FlowersGrassesFungi/   **
 **         meadowfoxtail.htm                                                **
 **                                                                          **
 ** Family name:         Gramineae                                           **
 ** Scientific name:     Alopecurus pratensis L.                             **
 ** Synonyms:                                                                **
 ** English name:        Meadow foxtail, Slender foxtail                     **
 ** Japanese name:       Ohsuzumenoteppoh                                    **
 ** Origin:              Europe, West Asia , North Africa                    **
 ** Introduction period: '64 - '80                                           **
 ** Life period:         Perennial                                           **
 ** Stem size:           50~100cm tall                                       **
 ** Stem:                Short creeping rhizomes. Culms erect, tufted.       **
 ** Leaf arrangement:    Alternate                                           **
 ** Leaves:              Leaf-blades 20~40 cm long, 3~5 mm wide, glabrous,   **
 **                      pale green; leaf-sheath pale green, often tinged    **
 **                      with purplish; ligules about 1~2 mm long.           **
 ** Blooming period:     May~July                                            **
 ** Inflorescence:       Spike                                               **
 ** Flower color:        pale green or tinged with purplish                  **
 ** Flower:              Inflorescence cylindrical, 5~8 cm long, 6~10 mm in  **
 **                      diameter, Spikelet 4~5 mm long, compressed,         **
 **                      1-floret; glumes equal as long as Spikelet,         **
 **                      3-nerved, folded by the main vein, long and soft    **
 **                      hairs on it; lemma 4-nerved, awns about 1 cm near   **
 **                      the rear base; anthers 2~3 mm long, dark yellow or  **
 **                      purple, protruding outside.                         **
 ** Fruits:              Caryopsis                                           **
 ** Seed:                                                                    **
 ** Distribution:        Europe, West Asia, North Africa                     **
 **     - http://alienplant.nier.go.kr/eng/html/search_detail.html?id=243    **
 **                                                                          **
 ** Images                                                                   **
 **     - http://forages.oregonstate.edu/main.cfm?PageID=178&specID=4#       **
 **                                                                          **
 ** TODO:                                                                    **
 **     - check the colour variations of the stem                            **
 **     - check the nature of the leaves                                     **
 **                                                                          **
 ******************************************************************************
 *****************************************************************************/



/*
 * Construct the stem of a Meadow Foxtail.
 *
 * i_rnd_seed: an initialized random number seed
 * i_stem_length: the length of the stem which the inflorescence is being added
 *     to
 */
#macro MakeMeadowFoxtailStem( i_rnd_seed, i_stem_length )
    #local STEM_RADIUS = 0.25;
    #local STEM_RADIUS_VARIANCE = 0.1;

    #local STEM_TEXTURE = pigment { rgb<76, 112, 38>/255 }

    object
    {
        MakeStraightStem(
            Random( i_rnd_seed, STEM_RADIUS, STEM_RADIUS_VARIANCE ),
            i_stem_length )
        texture { STEM_TEXTURE }
    }
#end


/*
 * Construct the inflorescence of a Meadow Foxtail.  If the global variable
 * RENDER_SPIKELETS is false the spikelets won't be created so as to speed
 * up rendering.
 *
 * i_rnd_seed: an initialized random number seed
 * i_stem_length: the length of the stem which the inflorescence is being added
 *     to
 */
#macro MakeMeadowFoxtailInflorescence( i_rnd_seed, i_stem_length )
    #local BODY_RADIUS = 0.8;
    #local BODY_RADIUS_VARIANCE = 0.2;

    #local BODY_LENGTH = 8;
    #local BODY_LENGTH_VARIANCE = 1;

    #local BODY_TEXTURE = pigment { rgb<76, 112, 38>/255 }

    #local body_radius =
        Random( i_rnd_seed, BODY_RADIUS, BODY_RADIUS_VARIANCE );
    #local body_length =
        Random( i_rnd_seed, BODY_LENGTH, BODY_LENGTH_VARIANCE );

    /*
     * Declare the body of the inflorescence.  We'll use this in a trace
     * further down the macro to add the spikelets if RENDER_SPIKELETS is
     * enabled, otherwise we'll just instanciate it immediately.
     */
    #local Body = sphere
    {
        0 body_radius
        texture { BODY_TEXTURE }

        scale <1, ( body_length / ( body_radius * 2 )), 1>
    }

    #if ( RENDER_SPIKELETS )
        #local SPIKELET_RADIUS = 0.01;
        #local SPIKELET_LENGTH = 0.5;
        #local SPIKELET_TEXTURE = pigment { rgb<219, 222, 145>/255 }

        #local SPIKELET_ANGLE = -30;
        #local SPIKELET_ANGLE_VARIANCE = 10;

        #local SPIKELET_CYLINDER_LENGTH = 0.5;
        #local SPIKELET_CYLINDER_RADIUS = 0.1;
        #local SPIKELET_CYLINDER_TEXTURE = pigment { rgb<230, 214, 142>/255 }

        #local SPIKELET_LENGTH_GAP = 0.5;
        #local SPIKELET_LENGTH_GAP_VARIANCE = 0.25;

        #local SPIKELET_RADIUS_GAP = 10;
        #local SPIKELET_RADIUS_GAP_VARIANCE = 5;

        /*
         * Declare the spikelet to be added at multiple points along the body
         * of the inflorescence.
         */
        #local Spikelet = union
        {
            object
            {
                MakeCurvedStem( SPIKELET_RADIUS, SPIKELET_LENGTH )
                texture { SPIKELET_TEXTURE }

                rotate <0, 180, -90>
            }
            cylinder
            {
                <SPIKELET_LENGTH, 0, 0>
                <SPIKELET_LENGTH, SPIKELET_CYLINDER_LENGTH, 0>
                SPIKELET_CYLINDER_RADIUS
                texture { SPIKELET_CYLINDER_TEXTURE }
            }
        }

        union
        {
            object
            {
                Body 
            }

            /*
             * Move along the height of the inflorescence adding spikelets as
             * we go.  length_idx is incremented by SPIKELET_LENGTH_GAP +/-
             * SPIKELET_LENGTH_GAP_VARIANCE each iteration.
             */
            #local length_idx = 0;

            #while ( length_idx <= body_length )
                #local target_loc =
                    <0, ( -( body_length * 0.5 ) + length_idx ), 0>;
                #local start_loc = <( body_radius + 0.1 ), target_loc.y, 0>;
                #local norm_loc = <0, 0, 0>; 
                #local intercept_loc = 
                    trace( Body, start_loc, ( target_loc - start_loc ),
                        norm_loc );

                #if ( vlength( norm_loc ) != 0 ) 
                    #local radius_idx = 0;

                    /*
                     * Move around the circumference of the inflorescence
                     * adding spikelets as we go.  radius_idx is incremented by
                     * SPIKELET_RADIUS_GAP +/- SPIKELET_RADIUS_GAP_VARIANCE
                     * each iteration.
                     */
                    #while ( radius_idx < 360 )
                        object
                        {
                            Spikelet
                            rotate <0,0,
                                Random( i_rnd_seed, SPIKELET_ANGLE,
                                    SPIKELET_ANGLE_VARIANCE )>
                            translate
                                ( intercept_loc - <0, SPIKELET_LENGTH, 0> )
                            rotate <0, radius_idx, 0>
                            scale ( 1 - ( length_idx * 0.01 ))
                        }

                        #local radius_idx =
                            ( radius_idx +
                                Random( i_rnd_seed, SPIKELET_RADIUS_GAP,
                                    SPIKELET_RADIUS_GAP_VARIANCE ));
                    #end
                #end

                #local length_idx =
                    ( length_idx +
                        Random( i_rnd_seed, SPIKELET_LENGTH_GAP,
                            SPIKELET_LENGTH_GAP_VARIANCE ));
            #end

            /*
             * Again move around the circumference of the inflorescence adding
             * spikelets as we go.  radius_idx is again incremented by
             * SPIKELET_RADIUS_GAP +/- * SPIKELET_RADIUS_GAP_VARIANCE each
             * iteration.  This is to remove the "bald" spot at the
             * inflorescence tip.  Rather than do traces here, we need to
             * slowly close the spikelets to form a cap over the tip by using
             * fairly specific values.
             */
            #local NUM_CAP_RINGS = 4;

            #local VERTICAL_ANGLES = array[ NUM_CAP_RINGS ];
            #local VERTICAL_ANGLES[ 0 ] = -20;
            #local VERTICAL_ANGLES[ 1 ] = -15;
            #local VERTICAL_ANGLES[ 2 ] = -2.5;
            #local VERTICAL_ANGLES[ 3 ] = 5;

            #local X_MOD = array[ NUM_CAP_RINGS ];
            #local X_MOD[ 0 ] = 0.4;
            #local X_MOD[ 1 ] = 0.3;
            #local X_MOD[ 2 ] = 0.2;
            #local X_MOD[ 3 ] = 0.1;

            #local Y_MOD = array[ NUM_CAP_RINGS ];
            #local Y_MOD[ 0 ] = 1.95;
            #local Y_MOD[ 1 ] = 1.8;
            #local Y_MOD[ 2 ] = 1.7;
            #local Y_MOD[ 3 ] = 1.6;

            #local cap_ring_idx = 0;

            #while ( cap_ring_idx < NUM_CAP_RINGS )
                #local radius_idx = 0;

                #while ( radius_idx < 360 )
                    object
                    {
                        Spikelet
                        rotate <0, 0,
                            Random( i_rnd_seed,
                                VERTICAL_ANGLES[ cap_ring_idx ],
                                SPIKELET_ANGLE_VARIANCE )>
                        translate <( body_radius * X_MOD[ cap_ring_idx ] ),
                            body_length/Y_MOD[ cap_ring_idx ], 0> -
                            <0, SPIKELET_LENGTH, 0>
                        rotate <0, radius_idx, 0>
                        scale 1 - ( length_idx * 0.01 )
                    }

                    #local radius_idx =
                        ( radius_idx +
                            Random( i_rnd_seed, SPIKELET_RADIUS_GAP,
                                SPIKELET_RADIUS_GAP_VARIANCE ) +
                            pow(( cap_ring_idx + 1 ), 2));
                #end

                #local cap_ring_idx = ( cap_ring_idx + 1 );
            #end

            rotate <0, 0, -15>
            translate StraightStemTipLocation( i_stem_length ) +
                <body_radius, ( body_length / 2 ), 0>
        }
    #else
        // The spikelets are not required, just instanciate the body
        object
        { 
            Body 
            rotate <0, 0, -15>
            translate StraightStemTipLocation( i_stem_length ) +
                <body_radius, ( body_length / 2 ), 0>
        }
    #end
#end


/*
 * Construct the leaves (between one and four) for a Meadow Foxtail.
 *
 * i_rnd_seed: an initialized random number seed
 * i_stem_length: the length of the stem which the leaves are being added to
 */
#macro MakeMeadowFoxtailLeaves( i_rnd_seed, i_stem_length )
    #local NUM_LEAVES = 2.5;
    #local NUM_LEAVES_VARIANCE = 1.5;

    #local num_leaves =
        floor( Random( i_rnd_seed, NUM_LEAVES, NUM_LEAVES_VARIANCE ));

    #if ( num_leaves != 0 )
        #local LEAF_BLADE_SEGMENT_HEIGHT = 0.25;

        #local LEAF_BLADE_LENGTH = 30;
        #local LEAF_BLADE_LENGTH_VARIANCE = 10;

        #local LEAF_BLADE_WIDTH = 0.4;
        #local LEAF_BLADE_WIDTH_VARIANCE = 0.1;

        #local MASTER_CURVE_SPEED = 0.05;
        #local MASTER_CURVE_SPEED_VARIANCE = 0.02;

        #local LENGTH_CURVE_SPEED = 0.005;
        #local LENGTH_CURVE_SPEED_VARIANCE = 0.002;

        #local PRE_CURVE_LENGTH = 0.6;
        #local PRE_CURVE_LENGTH_VARIANCE = 0.1;

        #local VERTICAL_CURVE_SPEED = 0.0005;
        #local VERTICAL_CURVE_SPEED_VARIANCE = 0.0003;

        #local LEAF_GAP_RADIANS =
            Radians( 360 / ( STRAIGHT_STEM_FRACTION * 12 ));

        #local NUM_LEAF_GAP_RADIANS = 5;
        #local NUM_LEAF_GAP_RADIANS_VARIANCE = 3;

        #local LEAF_TEXTURE = pigment { rgb<76, 112, 38>/255 }

        ConditionalUnion( num_leaves )
        {
            #local leaf_idx = 0;

            #while ( leaf_idx < num_leaves )
                object
                {
                    MakeLeafBlade( i_rnd_seed, LEAF_BLADE_SEGMENT_HEIGHT,
                        Random( i_rnd_seed, LEAF_BLADE_LENGTH,
                            LEAF_BLADE_LENGTH_VARIANCE ),
                        Random( i_rnd_seed, LEAF_BLADE_WIDTH,
                            LEAF_BLADE_WIDTH_VARIANCE ),
                        Random( i_rnd_seed, MASTER_CURVE_SPEED,
                            MASTER_CURVE_SPEED_VARIANCE ),
                        Random( i_rnd_seed, LENGTH_CURVE_SPEED,
                            LENGTH_CURVE_SPEED_VARIANCE ),
                        Random( i_rnd_seed, PRE_CURVE_LENGTH,
                            PRE_CURVE_LENGTH_VARIANCE ),
                        Random( i_rnd_seed, VERTICAL_CURVE_SPEED,
                            VERTICAL_CURVE_SPEED_VARIANCE ))

                    rotate <0, Random( i_rnd_seed, 180, 180 ), 0>
                    translate StraightStemLocation( i_stem_length,
                        ( LEAF_GAP_RADIANS *
                            Random( i_rnd_seed, NUM_LEAF_GAP_RADIANS,
                                NUM_LEAF_GAP_RADIANS_VARIANCE )))
                }

                #local leaf_idx = ( leaf_idx + 1 );
            #end

            texture { LEAF_TEXTURE }
        }
    #end
#end


/*
 * Construct a Meadow Foxtail.  If the global variable RENDER_INFLORESCENCE is
 * false the inflorescence won't be created so as to speed up rendering.  The
 * same goes for RENDER_LEAVES preventing the rendering of the leaves.
 *
 * i_rnd_seed: an initialized random number seed
 */
#macro MakeMeadowFoxtail( i_rnd_seed )
    #local STEM_LENGTH = 90;
    #local STEM_LENGTH_VARIANCE = 20;

    #local stem_length =
        Random( i_rnd_seed, STEM_LENGTH, STEM_LENGTH_VARIANCE );

    #if ( RENDER_INFLORESCENCE | RENDER_LEAVES )
        union
    #else
        object
    #end
    {
        MakeMeadowFoxtailStem( i_rnd_seed, stem_length )

        #if ( RENDER_INFLORESCENCE )
            MakeMeadowFoxtailInflorescence( i_rnd_seed, stem_length )
        #end

        #if ( RENDER_LEAVES )
            MakeMeadowFoxtailLeaves( i_rnd_seed, stem_length )
        #end
    }
#end


/*
 * Make a patch of a specified size of Meadow Foxtails centered at x=0, z=0
 * and resting on z=0, each one angled randomly.
 *
 * i_rnd_seed: an initialized random number seed
 * i_width: the width of the patch to constrain the meadow foxtails to
 * i_depth: the depth of the patch to constrain the meadow foxtails to
 * i_num_meadow_foxtails: the number of meadow foxtails to construct
 */
#macro MakeMeadowFoxtailPatch( i_rnd_seed, i_width, i_depth,
    i_num_meadow_foxtails )

    #local Z_ROTATION = 8.5;
    #local Z_ROTATION_VARIANCE = 5;

    #local Y_ROTATION = 180;
    #local Y_ROTATION_VARIANCE = 180;

    #if ( i_num_meadow_foxtails > 0 )
        ConditionalUnion( i_num_meadow_foxtails )
        {
            #local meadow_foxtail_idx = 0;

            #while ( meadow_foxtail_idx < i_num_meadow_foxtails )
                object
                {
                    MakeMeadowFoxtail( i_rnd_seed )

                    rotate <0, 0,
                        -( Random( i_rnd_seed, Z_ROTATION,
                            Z_ROTATION_VARIANCE ))>
                    rotate <0,
                        Random( i_rnd_seed, Y_ROTATION, Y_ROTATION_VARIANCE ),
                        0>
                    translate <(( rand( i_rnd_seed ) - 0.5 ) * i_width ), 0,
                        (( rand( i_rnd_seed ) - 0.5 ) * i_depth )>
                }

                #local meadow_foxtail_idx = ( meadow_foxtail_idx + 1 );
            #end
        }
    #end
#end
